from django.db.models.signals import pre_save
from django.dispatch import receiver
from arkindex.dataimport.models import WorkerRun
from rest_framework.exceptions import ValidationError


def _list_ancestors(graph, parents):
    results = []

    for parent_id in parents:
        results.append(parent_id)
        results.extend(_list_ancestors(graph, graph[parent_id]))

    return results


@receiver(pre_save, sender=WorkerRun)
def check_parents(sender, instance, **kwargs):
    """
    Since we can't properly validate parents field on WorkerRun by simply defining an Array of UUIDs,
    we add a pre-save signal with constraints.
    It will check that provided UUIDs refer to existing WorkerRun objects and that these objects are
    linked to the same DataImport as the to-be-saved WorkerRun instance.
    It will also check that adding or updating WorkerRun parents will not create cycles in the tree.
    """
    parents = WorkerRun.objects.filter(id__in=instance.parents, dataimport=instance.dataimport).exclude(id=instance.id).values_list('id', flat=True)
    if set(parents) != set(instance.parents):
        raise ValidationError(f"Can't add or update WorkerRun {instance.id} because parents field isn't properly defined. It can be either because"
                              " one or several UUIDs don't refer to existing WorkerRuns or either because listed WorkerRuns doesn't belong to the"
                              " same DataImport than this WorkerRun.")

    graph = {
        wr['id']: wr['parents']
        for wr in WorkerRun.objects.filter(dataimport=instance.dataimport).values('id', 'parents')
    }
    ancestors = _list_ancestors(graph, instance.parents)
    if instance.id in ancestors:
        raise ValidationError(f"Can't add or update WorkerRun {instance.id} because parents field isn't properly defined. It would create a cycle.")
