from unittest.mock import patch, MagicMock
from rest_framework.exceptions import APIException, NotAuthenticated, AuthenticationFailed, ValidationError
from gitlab.exceptions import GitlabGetError, GitlabCreateError
from arkindex.project.tests import FixtureTestCase
from arkindex.dataimport.providers import GitLabProvider
from arkindex.dataimport.models import DataImport, DataImportMode, Revision, WorkerVersionState, GitRefType


class TestGitLabProvider(FixtureTestCase):
    """
    Test the GitLabProvider class
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.creds = cls.user.credentials.get()
        cls.repo = cls.creds.repos.get()
        cls.rev = cls.repo.revisions.get()
        cls.gl_patch = patch('arkindex.dataimport.providers.Gitlab')

    def setUp(self):
        super().setUp()
        self.gl_mock = self.gl_patch.start()

    def tearDown(self):
        super().tearDown()
        self.gl_patch.stop()

    def test_list_repos(self):
        """
        Test GitLabProvider can list repositories from GitLab
        """
        GitLabProvider(url='http://aaa', credentials=self.creds).list_repos()

        self.assertEqual(self.gl_mock.call_count, 1)
        args, kwargs = self.gl_mock.call_args
        self.assertTupleEqual(args, ('http://aaa', ))
        self.assertDictEqual(kwargs, {'oauth_token': self.creds.token})

        self.assertEqual(self.gl_mock().projects.list.call_count, 1)
        args, kwargs = self.gl_mock().projects.list.call_args
        self.assertTupleEqual(args, ())
        self.assertDictEqual(kwargs, {'membership': True, 'search': None})

    def test_list_repos_query(self):
        """
        Test GitLabProvider can search repositories from GitLab
        """
        GitLabProvider(url='http://aaa', credentials=self.creds).list_repos(query='meh')

        self.assertEqual(self.gl_mock.call_count, 1)
        args, kwargs = self.gl_mock.call_args
        self.assertTupleEqual(args, ('http://aaa', ))
        self.assertDictEqual(kwargs, {'oauth_token': self.creds.token})

        self.assertEqual(self.gl_mock().projects.list.call_count, 1)
        args, kwargs = self.gl_mock().projects.list.call_args
        self.assertTupleEqual(args, ())
        self.assertDictEqual(kwargs, {'membership': True, 'search': 'meh'})

    def test_list_repos_requires_credentials(self):
        """
        Test GitLabProvider checks for credentials when requesting repositories list
        """
        with self.assertRaises(NotAuthenticated):
            GitLabProvider(url='http://aaa').list_repos()

    def test_create_repo(self):
        """
        Test GitLabProvider can create a Repository instance from a GitLab repo
        """
        self.gl_mock().projects.get.return_value.web_url = 'http://new_repo_url'
        self.gl_mock().projects.get.return_value.default_branch = 'branchname'

        request_mock = MagicMock()
        request_mock.build_absolute_uri.return_value = 'http://hook'
        new_repo = GitLabProvider(url='http://aaa', credentials=self.creds).create_repo(
            id='1337', request=request_mock, corpus=self.corpus)

        self.assertEqual(self.gl_mock().projects.get.call_count, 1)
        args, kwargs = self.gl_mock().projects.get.call_args
        self.assertTupleEqual(args, (1337, ))
        self.assertDictEqual(kwargs, {})

        self.assertEqual(new_repo.url, 'http://new_repo_url')
        self.assertEqual(new_repo.corpus, self.corpus)
        self.assertEqual(new_repo.provider_name, 'GitLabProvider')

        self.assertEqual(self.gl_mock().projects.get().hooks.create.call_count, 1)
        args, kwargs = self.gl_mock().projects.get().hooks.create.call_args
        self.assertEqual(len(args), 1)
        self.assertDictEqual(kwargs, {})
        self.assertDictEqual(args[0], {
            'url': 'http://hook',
            'push_events': True,
            'token': new_repo.hook_token,
        })

    def test_create_repo_requires_credentials(self):
        """
        Test GitLabProvider checks for credentials when requesting a repository creation
        """
        request_mock = MagicMock()
        request_mock.build_absolute_uri.return_value = 'http://hook'
        with self.assertRaises(NotAuthenticated):
            GitLabProvider(url='http://aaa').create_repo(
                id='repo_id', request=request_mock, corpus=self.corpus)

    def test_create_repo_already_exists(self):
        """
        Test GitLabProvider checks for duplicate repositories
        """
        self.gl_mock().projects.get.return_value.web_url = 'http://new_repo_url'
        self.gl_mock().projects.get.return_value.default_branch = 'branchname'

        request_mock = MagicMock()
        request_mock.build_absolute_uri.return_value = 'http://hook'
        GitLabProvider(url='http://aaa', credentials=self.creds).create_repo(
            id='1337', request=request_mock, corpus=self.corpus)

        with self.assertRaises(ValidationError):
            GitLabProvider(url='http://aaa', credentials=self.creds).create_repo(
                id='1337', request=request_mock, corpus=self.corpus)

    def test_create_repo_handle_get_error(self):
        """
        Test GitLabProvider handles GitLab repo GET errors
        """
        self.gl_mock().projects.get.side_effect = GitlabGetError

        request_mock = MagicMock()
        request_mock.build_absolute_uri.return_value = 'http://hook'

        with self.assertRaises(APIException):
            GitLabProvider(url='http://aaa', credentials=self.creds).create_repo(
                id='1337', request=request_mock, corpus=self.corpus)

        self.assertEqual(self.gl_mock().projects.get.call_count, 1)

    def test_create_repo_handle_hook_create_error(self):
        """
        Test GitLabProvider handles GitLab hook creation errors
        """
        self.gl_mock().projects.get.return_value.web_url = 'http://new_repo_url'
        self.gl_mock().projects.get.return_value.default_branch = 'branchname'
        self.gl_mock().projects.get.return_value.hooks.create.side_effect = GitlabCreateError

        request_mock = MagicMock()
        request_mock.build_absolute_uri.return_value = 'http://hook'

        with self.assertRaises(APIException):
            GitLabProvider(url='http://aaa', credentials=self.creds).create_repo(
                id='1337', request=request_mock, corpus=self.corpus)

        self.assertEqual(self.gl_mock().projects.get.call_count, 1)
        self.assertEqual(self.gl_mock().projects.get().hooks.create.call_count, 1)

    def test_update_or_create_ref(self):
        """
        Test GitLabProvider can create or update a GitRef instance for a repo and a revision
        """
        commit_refs = [
            {'name': 'refs/tags/v0.1.0', 'type': 'tag'},
            {'name': 'refs/heads/branch1', 'type': 'branch'},
            {'name': 'refs/heads/branch2', 'type': 'branch'},
        ]
        rev1 = Revision(
            repo=self.repo,
            hash='1',
            message='commit message',
            author='bob',
        )
        rev1.save()
        rev2 = Revision(
            repo=self.repo,
            hash='2',
            message='commit message',
            author='bob',
        )
        rev2.save()

        # Assert that git references are created properly
        for ref in commit_refs:
            GitLabProvider(url='http://aaa', credentials=self.creds) \
                .update_or_create_ref(self.repo, rev1, ref['name'], ref['type'])

        refs = [
            {'name': ref.name, 'type': ref.type, 'repo': ref.repository}
            for ref in rev1.refs.all()
        ]
        self.assertEqual(len(refs), 3)
        self.assertListEqual(refs, [
            {'name': 'refs/tags/v0.1.0', 'type': GitRefType.Tag, 'repo': self.repo},
            {'name': 'refs/heads/branch1', 'type': GitRefType.Branch, 'repo': self.repo},
            {'name': 'refs/heads/branch2', 'type': GitRefType.Branch, 'repo': self.repo},
        ])
        self.assertEqual(len(self.repo.refs.all()), 3)

        # Assert that git references are updated with another revision properly
        GitLabProvider(url='http://aaa', credentials=self.creds) \
            .update_or_create_ref(self.repo, rev2, commit_refs[0]['name'], commit_refs[0]['type'])
        GitLabProvider(url='http://aaa', credentials=self.creds) \
            .update_or_create_ref(self.repo, rev2, commit_refs[1]['name'], commit_refs[1]['type'])

        refs_rev1 = [
            {'name': ref.name, 'type': ref.type, 'repo': ref.repository}
            for ref in rev1.refs.all()
        ]
        refs_rev2 = [
            {'name': ref.name, 'type': ref.type, 'repo': ref.repository}
            for ref in rev2.refs.all()
        ]
        self.assertEqual(len(refs_rev1), 1)
        self.assertListEqual(refs_rev1, [
            {'name': 'refs/heads/branch2', 'type': GitRefType.Branch, 'repo': self.repo},
        ])
        self.assertEqual(len(refs_rev2), 2)
        self.assertListEqual(refs_rev2, [
            {'name': 'refs/tags/v0.1.0', 'type': GitRefType.Tag, 'repo': self.repo},
            {'name': 'refs/heads/branch1', 'type': GitRefType.Branch, 'repo': self.repo},
        ])
        self.assertEqual(len(self.repo.refs.all()), 3)

    def test_get_revision(self):
        """
        Test GitLabProvider can create a Revision instance for a repo by hash
        """
        revision, created = GitLabProvider(url='http://aaa', credentials=self.creds) \
            .get_or_create_revision(self.repo, '42')

        self.assertEqual(revision, self.rev)
        self.assertFalse(created)
        self.assertEqual(self.gl_mock.call_count, 0)

    def test_create_revision(self):
        """
        Test GitLabProvider can create a Revision instance for a repo by hash
        """
        self.gl_mock().projects.get.return_value.commits.get.return_value.refs.return_value = [
            {'name': 'refs/tags/v0.1.0', 'type': 'tag'},
            {'name': 'refs/heads/branch1', 'type': 'branch'},
            {'name': 'refs/heads/branch2', 'type': 'branch'},
            {'name': 'oh_no', 'type': 'all'},
        ]
        self.gl_mock().projects.get.return_value.commits.get.return_value.message = 'commit message'
        self.gl_mock().projects.get.return_value.commits.get.return_value.author_name = 'bob'

        revision, created = GitLabProvider(url='http://aaa', credentials=self.creds) \
            .get_or_create_revision(self.repo, '1337')

        self.assertTrue(created)
        self.assertEqual(revision.hash, '1337')

        # Assert that git references are created properly
        refs = [
            {
                'name': ref.name,
                'type': ref.type,
                'repo': ref.repository,
            }
            for ref in revision.refs.all()
        ]
        self.assertEqual(len(refs), 3)
        self.assertListEqual(refs, [
            {'name': 'refs/tags/v0.1.0', 'type': GitRefType.Tag, 'repo': self.repo},
            {'name': 'refs/heads/branch1', 'type': GitRefType.Branch, 'repo': self.repo},
            {'name': 'refs/heads/branch2', 'type': GitRefType.Branch, 'repo': self.repo},
        ])

        self.assertEqual(revision.message, 'commit message')
        self.assertEqual(revision.author, 'bob')

        self.assertEqual(self.gl_mock().projects.get.call_count, 1)
        self.assertEqual(self.gl_mock().projects.get().commits.get.call_count, 1)
        args, kwargs = self.gl_mock().projects.get().commits.get.call_args
        self.assertTupleEqual(args, ('1337', ))
        self.assertDictEqual(kwargs, {})

    def test_handle_webhook_missing_headers(self):
        """
        Test GitLabProvider checks HTTP headers on webhooks
        """
        glp = GitLabProvider(url='http://aaa', credentials=self.creds)

        request_mock = MagicMock()
        request_mock.data = {
            'object_kind': 'push',
            'ref': 'refs/heads/master',
            'checkout_sha': '1337',
            'commits': [
                {
                    'message': 'commit message',
                    'author': {
                        'name': 'bob',
                    }
                }
            ]
        }

        # Missing HTTP_X_GITLAB_TOKEN
        request_mock.META = {
            'HTTP_X_GITLAB_EVENT': 'Push Hook',
        }
        with self.assertRaises(NotAuthenticated):
            glp.handle_webhook(self.repo, request_mock)

        # Missing HTTP_X_GITLAB_EVENT
        request_mock.META = {
            'HTTP_X_GITLAB_TOKEN': 'hook-token',
        }
        with self.assertRaises(ValidationError):
            glp.handle_webhook(self.repo, request_mock)

        # Wrong HTTP_X_GITLAB_EVENT
        request_mock.META = {
            'HTTP_X_GITLAB_EVENT': 'Not a Push Hook',
            'HTTP_X_GITLAB_TOKEN': 'hook-token',
        }
        with self.assertRaises(ValidationError):
            glp.handle_webhook(self.repo, request_mock)

        # Wrong HTTP_X_GITLAB_TOKEN
        request_mock.META = {
            'HTTP_X_GITLAB_EVENT': 'Push Hook',
            'HTTP_X_GITLAB_TOKEN': 'not-the-hook-token',
        }
        with self.assertRaises(AuthenticationFailed):
            glp.handle_webhook(self.repo, request_mock)

    def test_handle_webhook_create_revision(self):
        """
        Test GitLabProvider does create a revision when a push event is received
        """
        request_mock = MagicMock()
        request_mock.META = {
            'HTTP_X_GITLAB_EVENT': 'Push Hook',
            'HTTP_X_GITLAB_TOKEN': 'hook-token',
        }
        sha = '1337'
        request_mock.data = {
            'object_kind': 'push',
            'ref': 'refs/heads/master',
            'checkout_sha': sha,
            'commits': [
                {
                    'message': 'commit message',
                    'author': {
                        'name': 'bob',
                    }
                }
            ]
        }

        self.gl_mock().projects.get.return_value.commits.get.return_value.refs.return_value = [
            {'name': 'oh_no', 'type': 'tag'},
            {'name': 'refs/heads/branch1', 'type': 'branch'},
            {'name': 'refs/heads/branch2', 'type': 'branch'},
        ]
        self.gl_mock().projects.get.return_value.commits.get.return_value.message = 'commit message'
        self.gl_mock().projects.get.return_value.commits.get.return_value.author_name = 'bob'

        rev = self.repo.revisions.filter(hash=sha)
        self.assertFalse(rev.exists())
        repo_imports = DataImport.objects.filter(revision__repo_id=str(self.repo.id))
        self.assertFalse(repo_imports.exists())
        GitLabProvider(url='http://aaa', credentials=self.creds).handle_webhook(self.repo, request_mock)

        revision = rev.get()
        self.assertEqual(revision.state, WorkerVersionState.Created)
        di = repo_imports.get()
        self.assertEqual(di.mode, DataImportMode.Repository)

    def test_handle_webhook_duplicate_events(self):
        """
        Test GitLabProvider checks for already handled events
        """
        request_mock = MagicMock()
        request_mock.META = {
            'HTTP_X_GITLAB_EVENT': 'Push Hook',
            'HTTP_X_GITLAB_TOKEN': 'hook-token',
        }
        sha = '42'
        request_mock.data = {
            'object_kind': 'push',
            'ref': 'refs/heads/master',
            'checkout_sha': sha,
            'commits': [
                {
                    'message': 'a',
                    'author': {
                        'name': 'me',
                    }
                }
            ]
        }

        rev = self.repo.revisions.filter(hash=sha)
        self.assertTrue(rev.exists())
        repo_imports = DataImport.objects.filter(revision__repo_id=str(self.repo.id))
        self.assertFalse(repo_imports.exists())
        GitLabProvider(url='http://aaa', credentials=self.creds).handle_webhook(self.repo, request_mock)

        # Checking that we didn't initiate revision creation
        self.assertEqual(self.gl_mock().projects.get.call_count, 0)
        self.assertEqual(self.gl_mock().projects.get().commits.get.call_count, 0)
        self.assertFalse(repo_imports.exists())

    def test_handle_webhook_wrong_kind(self):
        """
        Test GitLabProvider gracefully fails when GitLab breaks things
        """
        sha = '1337'
        rev = self.repo.revisions.filter(hash=sha)
        self.assertFalse(rev.exists())
        repo_imports = DataImport.objects.filter(revision__repo_id=str(self.repo.id))

        glp = GitLabProvider(url='http://aaa', credentials=self.creds)
        request_mock = MagicMock()
        request_mock.META = {
            'HTTP_X_GITLAB_EVENT': 'Push Hook',
            'HTTP_X_GITLAB_TOKEN': 'hook-token',
        }
        request_mock.data = {
            'object_kind': 'oh_no',  # Bad object kind
            'ref': 'refs/heads/master',
            'checkout_sha': sha,
            'commits': [
                {
                    'message': 'b',
                    'author': {
                        'name': 'me',
                    }
                }
            ]
        }
        with self.assertRaises(ValidationError):
            glp.handle_webhook(self.repo, request_mock)
        self.assertFalse(rev.exists())
        self.assertFalse(repo_imports.exists())

        # Missing SHA
        request_mock.data['object_kind'] = 'push'
        del request_mock.data['checkout_sha']
        with self.assertRaises(ValidationError):
            glp.handle_webhook(self.repo, request_mock)
        self.assertFalse(rev.exists())
        self.assertFalse(repo_imports.exists())

        # Breaking change: a list!
        request_mock.data = [request_mock.data]
        with self.assertRaises(ValidationError):
            glp.handle_webhook(self.repo, request_mock)
        self.assertFalse(rev.exists())
        self.assertFalse(repo_imports.exists())
