from arkindex_common.ml_tool import MLToolType
from arkindex_common.enums import DataImportMode
from arkindex.dataimport.models import Worker, WorkerVersion, WorkerRun
from arkindex.dataimport.signals import _list_ancestors
from arkindex.project.tests import FixtureAPITestCase
from rest_framework.exceptions import ValidationError


class TestSignals(FixtureAPITestCase):
    """
    Test signals
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.creds = cls.user.credentials.get()
        cls.repo = cls.creds.repos.get()
        cls.rev_1 = cls.repo.revisions.get()
        cls.rev_2 = cls.repo.revisions.create(
            hash='2',
            message='commit message',
            author='bob',
        )

        cls.worker_1 = Worker.objects.create(
            repository=cls.repo,
            name='Worker 1',
            slug='worker_1',
            type=MLToolType.Classifier
        )
        cls.version_1 = WorkerVersion.objects.create(
            worker=cls.worker_1,
            revision=cls.rev_1,
            configuration={"test": "test1"}
        )
        cls.version_2 = WorkerVersion.objects.create(
            worker=cls.worker_1,
            revision=cls.rev_2,
            configuration={"test": "test2"}
        )

        cls.dataimport_1 = cls.corpus.imports.create(creator=cls.user, mode=DataImportMode.Images)
        cls.run_1 = cls.dataimport_1.worker_runs.create(
            version=cls.version_1,
            parents=[],
        )
        cls.dataimport_2 = cls.corpus.imports.create(creator=cls.user, mode=DataImportMode.Images)

    def test_worker_run_check_parents_recursive(self):
        run_2 = self.dataimport_1.worker_runs.create(
            version=self.version_2,
            parents=[],
        )
        self.assertListEqual(run_2.parents, [])

        run_2.parents = [str(run_2.id)]
        with self.assertRaises(ValidationError):
            run_2.save()

        run_2.refresh_from_db()
        self.assertListEqual(run_2.parents, [])

    def test_worker_run_check_parents_distinct_dataimport(self):
        self.assertEqual(len(self.dataimport_1.worker_runs.all()), 1)
        with self.assertRaises(ValidationError):
            self.dataimport_2.worker_runs.create(
                version=self.version_2,
                parents=[str(self.run_1.id)],
            )
        self.dataimport_1.refresh_from_db()

        self.assertEqual(len(self.dataimport_1.worker_runs.all()), 1)

    def test_worker_run_check_parents_inexsistant_uuid(self):
        self.assertEqual(len(self.dataimport_1.worker_runs.all()), 1)
        with self.assertRaises(ValidationError):
            self.dataimport_2.worker_runs.create(
                version=self.version_2,
                parents=['12341234-1234-1234-1234-123412341234'],
            )
        self.dataimport_1.refresh_from_db()

        self.assertEqual(len(self.dataimport_1.worker_runs.all()), 1)

    def test_list_ancestors_no_parents(self):
        ancestors = _list_ancestors(WorkerRun.objects.none(), {})

        self.assertListEqual(ancestors, [])

    def test_list_ancestors_one_level(self):
        # --- TREE ---
        #
        #   B  C  D
        #    \ | /
        #     \|/
        #      A
        ancestors = _list_ancestors({"B": [], "C": [], "D": []}, ["B", "C", "D"])

        self.assertListEqual(ancestors, ["B", "C", "D"])

    def test_list_ancestors_three_level(self):
        # --- TREE ---
        #
        #   H     I  J
        #    \    | /
        #     \   |/
        #   E  F  G
        #    \ |  |
        #     \|  |
        #   B  C  D
        #    \ | /
        #     \|/
        #      A
        ancestors = _list_ancestors(
            {"B": [], "C": ["E", "F"], "D": ["G"], "E": [], "F": ["H"], "G": ["I", "J"], "H": [], "I": [], "J": []},
            ["B", "C", "D"]
        )

        self.assertListEqual(ancestors, ["B", "C", "E", "F", "H", "D", "G", "I", "J"])

    def test_worker_run_small_cycle(self):
        # --- TREE (unallowed cycle) ---
        #
        #             run_1
        #             |   |
        #             |   |
        #             run_2

        run_2 = self.dataimport_1.worker_runs.create(
            version=self.version_2,
            parents=[self.run_1.id],
        )

        self.run_1.parents = [run_2.id]
        with self.assertRaises(ValidationError):
            self.run_1.save()
        self.run_1.refresh_from_db()

        self.assertListEqual(self.run_1.parents, [])

    def test_worker_run_big_cycle(self):
        # --- TREE (unallowed cycle) ---
        #
        #             run_1
        #           /       \
        #        run_5     run_2
        #          |         |
        #           \       /
        #         run_4---run_3

        rev_3 = self.repo.revisions.create(
            hash='3',
            message='commit message',
            author='dylan',
        )
        rev_4 = self.repo.revisions.create(
            hash='4',
            message='commit message',
            author='pink',
        )
        rev_5 = self.repo.revisions.create(
            hash='5',
            message='commit message',
            author='floyd',
        )
        version_3 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_3,
            configuration={"test": "test3"}
        )
        version_4 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_4,
            configuration={"test": "test4"}
        )
        version_5 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_5,
            configuration={"test": "test5"}
        )

        run_2 = self.dataimport_1.worker_runs.create(
            version=self.version_2,
            parents=[self.run_1.id],
        )
        run_3 = self.dataimport_1.worker_runs.create(
            version=version_3,
            parents=[run_2.id],
        )
        run_4 = self.dataimport_1.worker_runs.create(
            version=version_4,
            parents=[run_3.id],
        )
        run_5 = self.dataimport_1.worker_runs.create(
            version=version_5,
            parents=[run_4.id],
        )

        self.run_1.parents = [run_5.id]
        with self.assertNumQueries(2), self.assertRaises(ValidationError):
            self.run_1.save()
        self.run_1.refresh_from_db()

        self.assertListEqual(self.run_1.parents, [])

    def test_worker_run_no_cycle(self):
        rev_3 = self.repo.revisions.create(
            hash='3',
            message='commit message',
            author='dylan',
        )
        version_3 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_3,
            configuration={"test": "test3"}
        )

        run_2 = self.dataimport_1.worker_runs.create(
            version=self.version_2,
            parents=[self.run_1.id],
        )
        run_3 = self.dataimport_1.worker_runs.create(
            version=version_3,
            parents=[run_2.id],
        )

        run_3.parents.append(self.run_1.id)
        with self.assertNumQueries(3):
            run_3.save()
        run_3.refresh_from_db()

        self.assertListEqual(run_3.parents, [run_2.id, self.run_1.id])

    def test_worker_run_check_parents(self):
        self.assertEqual(len(self.dataimport_1.worker_runs.all()), 1)
        run_2 = self.dataimport_1.worker_runs.create(
            version=self.version_2,
            parents=[self.run_1.id],
        )

        self.assertEqual(len(self.dataimport_1.worker_runs.all()), 2)
        self.assertListEqual(list(self.dataimport_1.worker_runs.all()), [
            self.run_1,
            run_2,
        ])
