from django.core.management.base import BaseCommand, CommandError
import sys
import os
import multiprocessing
from django.core.wsgi import get_wsgi_application

import logging

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s [%(levelname)s] %(message)s",
)
logger = logging.getLogger()


class Command(BaseCommand):
    help = "Run Arkindex API server through Gunicorn"

    def add_arguments(self, parser):
        parser.add_argument(
            "--host",
            type=str,
            help="Host to bind gunicorn",
            default=os.environ.get("HOST", "localhost"),
        )
        parser.add_argument(
            "--port",
            type=int,
            help="Port to bind gunicorn",
            default=int(os.environ.get("PORT", 8000)),
        )
        parser.add_argument(
            "--max-workers",
            help="Add a limit to the number of workers",
            type=int,
            default=int(os.environ.get("MAX_WORKERS", 0)),
        )

    def handle(self, host, port, max_workers, *args, **kwargs):
        try:
            from gunicorn.app.base import Application
        except ImportError:
            raise CommandError("Gunicorn is not available")

        # Calc max workers
        workers = (multiprocessing.cpu_count() * 2) + 1
        if max_workers > 0:
            workers = min(workers, max_workers)

        # Build bind string
        bind = f"{host}:{port}"
        logger.info(f"Running server on {bind} with {workers} workers")

        # Do not send out CLI args to gunicorn as they are not compatible
        # and we override directly the configuration to pass the WSGI application
        # This is needed to run through Nuitka
        sys.argv = sys.argv[:2]

        class ArkindexServer(Application):
            """Run the Django WSGI app through gunicorn"""
            def init(self, *args, **kwargs):
                return {
                    "bind": bind,
                    "workers": workers
                }

            def load(self):
                return get_wsgi_application()

        ArkindexServer().run()
