from django.urls import reverse
from django.test import override_settings
from unittest.mock import patch
from mock import AsyncMock
from rest_framework import status
from arkindex.project.tests import FixtureAPITestCase
from arkindex_common.enums import TranscriptionType
from arkindex_common.ml_tool import MLToolType
from arkindex.dataimport.models import Worker, WorkerVersion
from arkindex.documents.models import Corpus, Transcription, DataSource
from arkindex.users.models import User
from uuid import uuid4


class TestTranscriptionCreate(FixtureAPITestCase):
    """
    Tests for text element creation view
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.page = cls.corpus.elements.get(name='Volume 1, page 1r')
        cls.src = DataSource.objects.get(slug='test')
        cls.line = cls.corpus.elements.filter(type__slug='text_line').first()
        cls.private_corpus = Corpus.objects.create(name='Private')
        cls.private_page = cls.private_corpus.elements.create(type=cls.page.type)
        # Create an user with a read right only on the private corpus
        cls.private_read_user = User.objects.create_user('a@bc.de', 'a')
        cls.private_read_user.verified_email = True
        cls.private_read_user.save()
        cls.private_corpus.corpus_right.create(user=cls.private_read_user)
        cls.creds = cls.user.credentials.get()
        cls.repo = cls.creds.repos.get()
        cls.rev = cls.repo.revisions.get()

    def setUp(self):
        self.manual_source = DataSource.objects.create(type=MLToolType.Recognizer, slug='manual', internal=False)

    def test_create_transcription_require_login(self):
        response = self.client.post(reverse('api:transcription-create', kwargs={'pk': self.line.id}), format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {
            'detail': 'Authentication credentials were not provided.'
        })

    def test_create_transcription_write_right(self):
        self.client.force_login(self.private_read_user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.private_page.id}),
            format='json',
            data={
                'type': 'word',
                'text': 'NEKUDOTAYIM',
                'source': 'manual'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    @patch('arkindex.project.triggers.get_channel_layer')
    def test_create_transcription_no_element(self, get_layer_mock):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': uuid4()}),
            format='json',
            data={
                'type': 'word',
                'text': 'NEKUDOTAYIM',
                'source': 'manual'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    @patch('arkindex.project.triggers.get_channel_layer')
    def test_create_transcription_elt_null_zone(self, get_layer_mock):
        null_zone_page = self.corpus.elements.create(type=self.page.type)

        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': null_zone_page.id}),
            format='json',
            data={
                'type': 'word',
                'text': 'NEKUDOTAYIM',
                'source': 'manual'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'element': ['The element has no zone']
        })

    @patch('arkindex.project.triggers.get_channel_layer')
    def test_create_transcription(self, get_layer_mock):
        """
        Checks the view creates a manual transcription and runs ES indexing
        """
        get_layer_mock.return_value.send = AsyncMock()

        # The view must have the ability to create a manual source if it does not exist
        self.manual_source.delete()
        self.assertFalse(DataSource.objects.filter(type=MLToolType.Recognizer, slug='manual').exists())

        self.client.force_login(self.user)
        with self.assertNumQueries(10):
            response = self.client.post(
                reverse('api:transcription-create', kwargs={'pk': self.line.id}),
                format='json',
                data={
                    'type': 'line',
                    'text': 'A perfect day in a perfect place',
                    'source': 'manual'
                }
            )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        tr = Transcription.objects.get(text='A perfect day in a perfect place')
        self.assertDictEqual(response.json(), {
            'id': str(tr.id),
            'score': None,
            'source': {
                'id': str(tr.source.id),
                'internal': False,
                'name': '',
                'revision': '',
                'slug': 'manual',
                'type': 'recognizer'
            },
            'text': 'A perfect day in a perfect place',
            'type': 'line',
            'zone': None
        })

        new_ts = Transcription.objects.get(text='A perfect day in a perfect place', type=TranscriptionType.Line)
        self.assertIsNone(new_ts.zone)
        self.assertIsNone(new_ts.score)
        self.assertTrue(DataSource.objects.filter(type=MLToolType.Recognizer, slug='manual').exists())
        self.assertEqual(new_ts.source.slug, 'manual')
        self.assertTrue(self.line.transcriptions.filter(pk=new_ts.id).exists())

        get_layer_mock().send.assert_called_once_with('reindex', {
            'type': 'reindex.start',
            'element': str(self.line.id),
            'corpus': None,
            'entity': None,
            'transcriptions': True,
            'elements': True,
            'entities': False,
            'drop': False,
        })

    @patch('arkindex.project.triggers.get_channel_layer')
    def test_create_transcription_takes_no_zone(self, get_layer_mock):
        """
        TranscriptionCreate endpoint do not create any zone for the transcription
        """
        get_layer_mock.return_value.send = AsyncMock()

        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={
                'type': 'line',
                'polygon': [(0, 0), (42, 0), (42, 42), (0, 42), (0, 0)],
                'text': 'SQUARE',
                'source': 'manual'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        new_ts = Transcription.objects.get(text='SQUARE', type=TranscriptionType.Line)
        self.assertEqual(new_ts.zone, None)

    @patch('arkindex.project.triggers.get_channel_layer')
    def test_create_duplicated_transcription(self, get_layer_mock):
        """
        Check the view creates a new transcriptions with a similar manual source
        """
        get_layer_mock.return_value.send = AsyncMock()

        self.client.force_login(self.user)
        ts = self.page.transcriptions.create(
            text='GLOUBIBOULGA',
            type=TranscriptionType.Word,
            source=self.manual_source
        )
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse('api:transcription-create', kwargs={'pk': self.line.id}),
                format='json',
                data={
                    'type': 'line',
                    'text': ts.text,
                    'source': 'manual'
                }
            )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(Transcription.objects.filter(text=ts.text, source=self.manual_source).count(), 2)
        ts.refresh_from_db()
        self.assertNotEqual(ts.score, 0.99)

    @patch('arkindex.project.triggers.get_channel_layer')
    def test_create_transcription_wrong_type(self, get_layer_mock):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={
                'type': 'AAAAA',
                'text': 'NEKUDOTAYIM',
                'source': 'manual'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'type': ['Value is not of type TranscriptionType']
        })

    @patch('arkindex.project.triggers.get_channel_layer')
    @override_settings(ARKINDEX_FEATURES={'search': False})
    def test_create_transcription_no_search(self, get_layer_mock):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={
                'type': 'line',
                'text': 'A classy text line',
                'source': 'manual'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertFalse(get_layer_mock().send.called)

    def test_create_transcription_no_source_no_worker_version(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={
                'type': 'word',
                'text': 'NEKUDOTAYIM',
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'source': ['This field XOR worker_version field must be set to create a transcription'],
            'worker_version': ['This field XOR source field must be set to create a transcription']
        })

    def test_create_transcription_source_and_worker_version_returns_error(self):
        self.client.force_login(self.user)
        worker = Worker.objects.create(
            repository=self.repo,
            name='Worker 1',
            slug='worker_1',
            type=MLToolType.Classifier
        )
        version = WorkerVersion.objects.create(
            worker=worker,
            revision=self.rev,
            configuration={"test": "test1"}
        )
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={
                'type': 'word',
                'text': 'NEKUDOTAYIM',
                'source': 'manual',
                'worker_version': str(version.id),
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'source': ['You can only refer to a DataSource XOR a WorkerVersion on a transcription'],
            'worker_version': ['You can only refer to a DataSource XOR a WorkerVersion on a transcription']
        })

    def test_create_transcription_worker_version_non_internal(self):
        self.client.force_login(self.user)
        worker = Worker.objects.create(
            repository=self.repo,
            name='Worker 1',
            slug='worker_1',
            type=MLToolType.Classifier
        )
        version = WorkerVersion.objects.create(
            worker=worker,
            revision=self.rev,
            configuration={"test": "test1"}
        )
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={
                'type': 'word',
                'text': 'NEKUDOTAYIM',
                'worker_version': str(version.id),
                'score': .42
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'worker_version': [f'An internal user is required to create a transcription with the worker_version "{version.id}"']
        })

    @patch('arkindex.project.triggers.get_channel_layer')
    def test_create_transcription_worker_version(self, get_layer_mock):
        get_layer_mock.return_value.send = AsyncMock()

        self.client.force_login(self.internal_user)
        worker = Worker.objects.create(
            repository=self.repo,
            name='Worker 1',
            slug='worker_1',
            type=MLToolType.Classifier
        )
        version = WorkerVersion.objects.create(
            worker=worker,
            revision=self.rev,
            configuration={"test": "test1"}
        )
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={
                'type': 'word',
                'text': 'NEKUDOTAYIM',
                'worker_version': str(version.id),
                'score': .42
            }
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        tr = Transcription.objects.get(text='NEKUDOTAYIM')
        self.assertEqual(tr.worker_version, version)
        self.assertDictEqual(response.json(), {
            'id': str(tr.id),
            'score': .42,
            'source': None,
            'text': 'NEKUDOTAYIM',
            'type': 'word',
            'zone': None
        })

        get_layer_mock().send.assert_called_once_with('reindex', {
            'type': 'reindex.start',
            'element': str(self.line.id),
            'corpus': None,
            'entity': None,
            'transcriptions': True,
            'elements': True,
            'entities': False,
            'drop': False,
        })

    def test_manual_transcription_forbidden_type(self):
        """
        Creating a manual transcription with a non allowed type is forbidden
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={
                'type': 'word',
                'text': 'NEKUDOTAYIM',
                'source': 'manual'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'type': ["Only transcriptions of type 'line' are allowed for this element"]
        })

    def test_manual_transcription_no_allowed_type(self):
        self.line.type.allowed_transcription = None
        self.line.type.save()
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={
                'type': 'line',
                'text': 'A classy text line',
                'source': 'manual'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'element': ['The element type does not allow creating a manual transcription']
        })

    @patch('arkindex.project.serializer_fields.MLTool.get')
    def test_ml_transcription_non_internal(self, ml_get_mock):
        """
        A transcription created on an element with an internal source
        should only be authorized for an internal user (e.g. a worker)
        """
        ml_get_mock.return_value.type = self.src.type
        ml_get_mock.return_value.slug = self.src.slug
        ml_get_mock.return_value.name = self.src.name
        ml_get_mock.return_value.version = self.src.revision
        self.src.internal = True
        self.src.save()

        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={
                'type': 'word',
                'text': 'SQUARE',
                'source': self.src.slug,
                'score': .99
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'source': [
            'An internal user is required to create a transcription with the internal source "test"']
        })

    @patch('arkindex.project.serializer_fields.MLTool.get')
    def test_ml_trancription_required_score(self, ml_get_mock):
        """
        A score is required when creating a transcription on an
        element with an internal source
        """
        ml_get_mock.return_value.type = self.src.type
        ml_get_mock.return_value.slug = self.src.slug
        ml_get_mock.return_value.name = self.src.name
        ml_get_mock.return_value.version = self.src.revision
        self.src.internal = True
        self.src.save()

        self.client.force_login(self.internal_user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={
                'type': 'word',
                'text': 'CIRCLE',
                'source': self.src.slug
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'score': ['This field is required for non manual sources.']})

    @patch('arkindex.project.serializer_fields.MLTool.get')
    def test_ml_trancription_requires_element_zone(self, ml_get_mock):
        """
        A worker can publish a transcription on an element with a zone only
        """
        null_zone_page = self.corpus.elements.create(type=self.page.type)
        ml_get_mock.return_value.type = self.src.type
        ml_get_mock.return_value.slug = self.src.slug
        ml_get_mock.return_value.name = self.src.name
        ml_get_mock.return_value.version = self.src.revision
        self.src.internal = True
        self.src.save()

        self.client.force_login(self.internal_user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': null_zone_page.id}),
            format='json',
            data={
                'type': 'word',
                'text': 'ELLIPSE',
                'source': self.src.slug,
                'score': .42
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'element': ['The element has no zone']})

    @patch('arkindex.project.serializer_fields.MLTool.get')
    @patch('arkindex.project.triggers.get_channel_layer')
    def test_ml_transcription_create(self, get_layer_mock, ml_get_mock):
        """
        Creates a transcription depending on an element zone
        """
        get_layer_mock.return_value.send = AsyncMock()
        ml_get_mock.return_value.type = self.src.type
        ml_get_mock.return_value.slug = self.src.slug
        ml_get_mock.return_value.name = self.src.name
        ml_get_mock.return_value.version = self.src.revision
        self.src.internal = True
        self.src.save()

        self.client.force_login(self.internal_user)
        self.assertTrue(self.internal_user.is_internal)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={
                'type': 'word',
                'text': 'TRIANGLE',
                'source': self.src.slug,
                'score': .42
            }
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(self.line.transcriptions.filter(source=self.src).count(), 1)

        get_layer_mock().send.assert_called_once_with('reindex', {
            'type': 'reindex.start',
            'element': str(self.line.id),
            'corpus': None,
            'entity': None,
            'transcriptions': True,
            'elements': True,
            'entities': False,
            'drop': False,
        })
