from django.urls import reverse
from rest_framework import status
from arkindex_common.enums import MetaType, EntityType
from arkindex.documents.models import \
    DataSource, Corpus, Entity, MLClass, Classification
from arkindex.project.tests import FixtureAPITestCase


class TestRetrieveElements(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.vol = cls.corpus.elements.get(name='Volume 1')
        cls.private_corpus = Corpus.objects.create(name='private', public=False)

    def setUp(self):
        self.page = self.corpus.elements.get(name='Volume 1, page 1r')
        self.folio_metadata = self.page.metadatas.get()
        self.metadata = self.page.metadatas.create(
            name='meta',
            type=MetaType.Text,
            value='text_metadata'
        )

    def test_get_element(self):
        data_source = DataSource.objects.get(slug='test')
        ml_class = MLClass.objects.create(name='text', corpus=self.corpus)
        classification = Classification.objects.create(element=self.vol, source=data_source, ml_class=ml_class)

        response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['id'], str(self.vol.id))
        self.assertEqual(data['name'], 'Volume 1')
        self.assertEqual(data['corpus']['id'], str(self.corpus.id))
        self.assertEqual(data['type'], 'volume')
        self.assertEqual(data['thumbnail_url'], self.vol.thumbnail.s3_url)
        self.assertIsNone(data['thumbnail_put_url'])
        self.assertIsNone(data['source'])

        self.assertEqual(len(data['classifications']), 1)
        cl = data['classifications'][0]
        self.assertEqual(cl['id'], str(classification.id))
        self.assertEqual(cl['source']['id'], str(classification.source.id))
        self.assertEqual(cl['ml_class']['id'], str(classification.ml_class.id))

    def test_get_element_thumbnail_folder(self):
        """
        Check getting an element only gives a thumbnail URL with folders
        """
        self.assertTrue(self.vol.type.folder)
        response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['thumbnail_url'], self.vol.thumbnail.s3_url)
        self.assertIsNone(data['thumbnail_put_url'])

        self.assertFalse(self.page.type.folder)
        response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertIsNone(data['thumbnail_url'])
        self.assertIsNone(data['thumbnail_put_url'])

    def test_get_element_thumbnail_put_internal(self):
        """
        Check getting an element only gives a thumbnail PUT URL to bots for folders
        """
        self.client.force_login(self.internal_user)

        self.assertTrue(self.vol.type.folder)
        response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['thumbnail_put_url'], self.vol.thumbnail.s3_put_url)

        self.assertFalse(self.page.type.folder)
        response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIsNone(response.json()['thumbnail_put_url'])

    def test_non_empty_folder(self):
        self.client.force_login(self.user)
        response = self.client.delete(reverse('api:element-retrieve', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'element': "Element '{}' is linked to 15 children elements".format(self.vol.id)}
        )

    def test_list_element_metadata(self):
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['metadata']
        self.assertListEqual(results, [
            {
                'dates': [],
                'id': str(self.folio_metadata.id),
                'name': 'folio',
                'revision': None,
                'type': 'text',
                'value': '1r',
            },
            {
                'id': str(self.metadata.id),
                'name': str(self.metadata.name),
                'type': self.metadata.type.value,
                'value': self.metadata.value,
                'dates': [],
                'revision': None
            },
        ])

    def test_list_element_metadata_with_entity(self):
        # Link an entity to metadata, db requests number should remains same
        entity = Entity.objects.create(
            type=EntityType.Person,
            corpus=self.corpus,
            name='Marc',
            source_id=DataSource.objects.get(slug='test').id
        )
        self.metadata.entity = entity
        self.metadata.save()
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['metadata']
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]['name'], 'folio')

    def test_list_element_metadata_wrong_acl(self):
        self.page.corpus = self.private_corpus
        self.page.save()
        response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
