from pathlib import Path
from django.core.management.base import BaseCommand, CommandError
from arkindex.project.argparse import CorpusArgument
from arkindex.documents.models import ElementType
from arkindex.himanis.tei import TeiParser
import logging

logging.basicConfig(
    format='%(asctime)s [%(levelname)s] %(message)s',
    level=logging.INFO,
)


class Command(BaseCommand):
    help = "Import act metadata from TEI-XML files"

    def add_arguments(self, parser):
        parser.add_argument(
            'xml',
            help='Path to a folder containing XML files with act metadata.',
            type=Path,
        )
        parser.add_argument(
            '--corpus',
            type=CorpusArgument(),
            help='ID or part of the name of a corpus to import acts and surfaces into.',
            required=True,
        )
        parser.add_argument(
            '--folder-type',
            dest='folder_slug',
            help='Slug of an element type for folders in the selected corpus.',
            default='volume',
        )
        parser.add_argument(
            '--act-type',
            dest='act_slug',
            help='Slug of an element type for acts in the selected corpus.',
            default='act',
        )

    def handle(self, xml, corpus, folder_slug, act_slug, **options):
        if not xml.is_dir():
            raise CommandError('Path {!s} does not exist'.format(xml))

        try:
            folder_type = corpus.types.get(slug=folder_slug, folder=True)
        except ElementType.DoesNotExist:
            raise CommandError(
                'Folder type with slug {!r} in corpus {!s} does not exist'.format(folder_slug, corpus.name)
            )

        try:
            act_type = corpus.types.get(slug=act_slug, folder=False)
        except ElementType.DoesNotExist:
            raise CommandError(
                'Act type with slug {!r} in corpus {!s} does not exist'.format(act_slug, corpus.name)
            )

        for path in xml.rglob('*.xml'):
            try:
                parser = TeiParser(path, corpus, folder_type, act_type)
                parser.run()
            except Exception as e:
                self.stdout.write(self.style.ERROR(str(e)))
