from collections import namedtuple
from collections.abc import Iterable, Sequence
from django.db import models
from django.utils.functional import cached_property
import math
import re

REGEX_SPLIT = re.compile(r'\(([0-9-]+),\s?([0-9-]+)\)')


class Point(namedtuple('Point', ['x', 'y'])):
    """
    Describe a point in the space.
    """
    __slots__ = ()

    def __new__(cls, x, y=None):
        # Allow both Point(1, 2) and Point(tuple(1, 2)) for Django 3 compatibility
        if isinstance(x, Iterable):
            x, y = x
        return super().__new__(cls, int(x), int(y))

    def __str__(self):
        return '({0.x},{0.y})'.format(self)

    def __lt__(self, other):
        if isinstance(other, self.__class__):
            return self.x < other.x or (self.x == other.x and self.y < other.y)

        elif isinstance(other, tuple):
            if len(tuple) != 2:
                return NotImplemented
            return self.x < other[0] or (self.x == other[0] and self.y < other[1])

        return NotImplemented

    def __gt__(self, other):
        if isinstance(other, self.__class__):
            return self.x > other.x or (self.x == other.x and self.y > other.y)

        elif isinstance(other, tuple):
            if len(tuple) != 2:
                return NotImplemented
            return self.x > other[0] or (self.x == other[0] and self.y > other[1])

        return NotImplemented

    def __mul__(self, other):
        if isinstance(other, (int, float)):
            return Point(self.x * other, self.y * other)
        return NotImplemented

    def angle_relative_to(self, origin):
        return math.atan2(
            origin.y - self.y,
            origin.x - self.x,
        )


class Polygon(list):
    '''
    A hashable Polygon in-memory
    '''
    def __init__(self, points, reorder=True):
        super().__init__(points)
        assert len(self) > 1
        if not all(isinstance(point, Point) for point in self):
            for i in range(len(self)):
                self[i] = Point(*self[i])

        if reorder:
            self.reorder()

    @staticmethod
    def from_dict(data, reorder=True):
        '''
        Helper to build an instance from dict data:
        * from a direct polygon as list of tuples
        * from x,y + width,height
        '''
        if isinstance(data, list):
            return Polygon(data, reorder=reorder)
        assert isinstance(data, dict)

        # Direct usage
        if 'polygon' in data:
            if isinstance(data['polygon'], list):
                return Polygon(data['polygon'], reorder=reorder)
            elif isinstance(data['polygon'], Polygon):
                return data['polygon']
            else:
                raise ValueError('Unsupported value for polygon: {}'.format(data['polygon']))

        # Build from coords
        for k in ('x', 'y', 'width', 'height'):
            assert k in data, 'Missing key {}'.format(k)
            assert isinstance(data[k], (int, float)), 'Value of {} must be int or float'.format(k)

        return Polygon([
            (data['x'], data['y']),
            (data['x'] + data['width'], data['y']),
            (data['x'] + data['width'], data['y'] + data['height']),
            (data['x'], data['y'] + data['height']),
            (data['x'], data['y']),
        ])

    @staticmethod
    def from_str(data, reorder=True):
        assert isinstance(data, str)
        return Polygon([
            Point(float(x), float(y))
            for x, y in REGEX_SPLIT.findall(data)
        ], reorder=reorder)

    @staticmethod
    def from_coords(x, y, width, height, reorder=True):
        return Polygon.from_dict({'x': x, 'y': y, 'width': width, 'height': height}, reorder=reorder)

    def __repr__(self):
        return 'Polygon({!r})'.format(list(map(tuple, self)))

    def __eq__(self, other):
        if not isinstance(other, Sequence):
            return NotImplemented

        if len(self) != len(other):
            return False
        return all(
            x == y
            for x, y in zip(self, other)
        )

    def __hash__(self):
        return hash(tuple(self))

    def __mul__(self, other):
        if isinstance(other, (int, float)):
            return Polygon([point * other for point in self])
        return NotImplemented

    def serialize(self, as_string=False):
        '''
        Used to save data in DB
        '''
        if as_string is True:
            return '({})'.format(','.join(map(str, self)))
        return list(map(tuple, self))

    def reorder(self):
        """
        Ensure polygons are unique, mathematically speaking:
        - Remove duplicate points (ABBC → ABC)
        - Start the points from the point closest to (0, 0)
        - Make the points run clockwise (ABC == ACB)
        - Add PostgreSQL's extra point (ABC → ABCA)
        """
        # Deduplicate points: ABBCCBCA → ABCBCA
        # Use a while loop here instead of for, to remove points while iterating.
        # Ensure the polygon still has at least 3 points
        i = 0
        while len(self) > 3 and i < len(self):
            if self[i] == self[i - 1]:
                del self[i]
            else:
                i += 1

        # Use the point closest to (0, 0) as the origin point
        origin = min(self)
        origin_index = self.index(origin)

        forward_neighbor = self[(origin_index + 1) % len(self)]
        backward_neighbor = self[origin_index - 1]

        # If both neighbors are the same point, we cannot guess in which direction to go
        assert forward_neighbor != backward_neighbor, \
            'Cannot reorder this polygon: the origin point is surrounded by twice the same point.'

        # Compute the angles between the origin and its two neighboring points:
        # the smallest angle will tell in which order to go to have the points run clockwise
        forward_angle = forward_neighbor.angle_relative_to(origin)
        backward_angle = backward_neighbor.angle_relative_to(origin)

        # The forward neighbor's angle is greater: reverse all the points
        # There is an edge case when both neighbors have the same angle,
        # for example with a triangle where all points are on the same line.
        # We handle it by comparing the points themselves
        if forward_angle > backward_angle \
                or (forward_angle == backward_angle and forward_neighbor > backward_neighbor):
            self.reverse()
            # Grab the new index of the origin point
            origin_index = self.index(origin)

        if origin_index != 0:
            # Shift all the points to turn the origin point into the first point
            for i, point in enumerate(self[origin_index:] + self[:origin_index]):
                self[i] = point

        # Append PostgreSQL's closing point
        if self[0] != self[-1]:
            self.append(self[0])

    @cached_property
    def x(self):
        return min([point.x for point in self])

    @cached_property
    def y(self):
        return min([point.y for point in self])

    @cached_property
    def width(self):
        return max([point.x for point in self]) - self.x

    @cached_property
    def height(self):
        return max([point.y for point in self]) - self.y

    @property
    def center(self):
        return Point(
            self.x + self.width / 2.0,
            self.y + self.height / 2.0,
        )


class PolygonField(models.Field):
    """
    Field to store a polygon; needs at least three set of points
    """
    def db_type(self, connection):
        return 'polygon'

    def to_python(self, values):
        if values is None:
            return None

        if isinstance(values, str):
            polygon = Polygon.from_str(values)

        if not isinstance(values, Iterable):
            raise TypeError("Value {} is not iterable".format(values))

        return polygon

    def from_db_value(self, value, expression, connection):
        if value is None:
            return

        return Polygon.from_str(value, reorder=False)

    def get_prep_value(self, values):
        if values is None:
            return
        if isinstance(values, list) or isinstance(values, tuple):
            if len(values) < 3:
                raise ValueError("Needs at minimum 3 points")
            polygon = Polygon(values)
        elif isinstance(values, str):
            polygon = Polygon.from_str(values)
        elif isinstance(values, Polygon):
            polygon = values
        else:
            raise Exception('Not supported polygon input')

        return ', '.join(map(str, polygon)) if values else None

    def get_prep_lookup(self, lookup_type, value):
        raise NotImplementedError(self)


@PolygonField.register_lookup
class PolygonContains(models.Lookup):
    lookup_name = 'contains'

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = lhs_params + rhs_params
        return "%s @> %s::polygon" % (lhs, rhs), params


@PolygonField.register_lookup
class PolygonIn(models.Lookup):
    lookup_name = 'in'

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = lhs_params + rhs_params
        return "%s <@ %s::polygon" % (lhs, rhs), params


@PolygonField.register_lookup
class PolygonExact(models.Lookup):
    lookup_name = 'exact'

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = lhs_params + rhs_params
        return "%s ~= %s::polygon" % (lhs, rhs), params
