from unittest import TestCase
from arkindex.project.polygon import Polygon, Point


class TestPolygon(TestCase):
    '''
    Test the low level class
    '''

    @classmethod
    def setUpClass(cls):
        """
          x 0     200
        y 0 A-------B
            |       |
        100 D-------C
        """
        super().setUpClass()
        cls.a = Point(0, 0)
        cls.b = Point(0, 100)
        cls.c = Point(200, 100)
        cls.d = Point(200, 0)

    def test_base(self):
        polygon = Polygon([
            [0, 0],
            [0, 100],
            [200, 100],
            [200, 0],
        ])
        self.assertEqual(str(polygon), 'Polygon([(0, 0), (0, 100), (200, 100), (200, 0), (0, 0)])')
        self.assertEqual(polygon, Polygon.from_coords(0, 0, 200, 100))
        self.assertEqual(polygon.x, 0)
        self.assertEqual(polygon.y, 0)
        self.assertEqual(polygon.width, 200)
        self.assertEqual(polygon.height, 100)
        self.assertEqual(polygon.center.x, 100)
        self.assertEqual(polygon.center.y, 50)
        self.assertListEqual(polygon, [
            self.a, self.b, self.c, self.d, self.a
        ])

    def test_dict(self):
        polygon = Polygon.from_dict({
            'dummy': 'test',
            'polygon': [[0, 0], [0, 100], [200, 100], [200, 0]],
        })
        self.assertEqual(polygon.x, 0)
        self.assertEqual(polygon.y, 0)
        self.assertEqual(polygon.width, 200)
        self.assertEqual(polygon.height, 100)
        self.assertListEqual(polygon, [
            self.a, self.b, self.c, self.d, self.a
        ])

        polygon = Polygon.from_dict({
            'dummy': 'anything',
            'x': 0,
            'y': 0,
            'width': 200,
            'height': 100,
        })
        self.assertEqual(polygon.x, 0)
        self.assertEqual(polygon.y, 0)
        self.assertEqual(polygon.width, 200)
        self.assertEqual(polygon.height, 100)
        self.assertListEqual(polygon, [
            self.a, self.b, self.c, self.d, self.a
        ])

    def test_str(self):
        polygon = Polygon.from_str('((0, 0), (0,100),(200, 100),(200,0))')
        self.assertEqual(len(polygon), 5)
        self.assertEqual(polygon.x, 0)
        self.assertEqual(polygon.y, 0)
        self.assertEqual(polygon.width, 200)
        self.assertEqual(polygon.height, 100)

    def test_reorder(self):
        # BADCB → ABCDA
        polygon = Polygon([self.b, self.a, self.d, self.c, self.b])
        self.assertListEqual(polygon, [self.a, self.b, self.c, self.d, self.a])

    def test_no_reorder(self):
        polygon = Polygon([self.b, self.a, self.d, self.c, self.b], reorder=False)
        self.assertListEqual(polygon, [self.b, self.a, self.d, self.c, self.b])

    def test_duplicate_points(self):
        polygon = Polygon([
            self.a,
            self.a,
            self.b,
            self.b,
            self.c,
            self.c,
            self.c,
            self.b,
            self.d,
            self.a,
        ])
        # AABBCCCBDA → ABCBDA
        self.assertListEqual(polygon, [
            self.a, self.b, self.c, self.b, self.d, self.a,
        ])

    def test_closing_point(self):
        """
        Ensure the reordering adds the closing point PostgreSQL wants
        """
        polygon = Polygon([self.a, self.b, self.c, self.d])
        self.assertListEqual(polygon, [self.a, self.b, self.c, self.d, self.a])

    def test_scaling(self):
        polygon = Polygon([(0, 0), (0, 10), (10, 10), (10, 0), (0, 0)])
        self.assertListEqual(polygon * 2.5, [(0, 0), (0, 25), (25, 25), (25, 0), (0, 0)])
        self.assertListEqual(polygon * 0.5, [(0, 0), (0, 5), (5, 5), (5, 0), (0, 0)])
        # No change in the original polygon after any operation
        self.assertListEqual(polygon, [(0, 0), (0, 10), (10, 10), (10, 0), (0, 0)])
        with self.assertRaises(TypeError):
            polygon * polygon
