from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('documents', '0008_remove_regionelement'),
    ]

    operations = [
        # This performs the whole regions migration in just a few SQL queries
        # by exploiting the fact that UUIDs have a lot of combinations and it is
        # reasonable to assume that they are unique even across tables.

        # This turns MLClass IDs into element type IDs and region IDs into element IDs.
        # This way, no checks are required to ensure the foreign key constraints are respected.
        # The element names are generated as the class name and a number grouped by parent element.
        # Then, element paths are created in two steps:
        # * Create the paths from scratch for regions whose elements do not have paths;
        # * Create the paths for regions whose elements already have paths by prepending the element ID to all paths.
        # We do not have to care about existing paths, cycles or existing orderings since
        # everything is in a transaction and uses new types.

        # The uuid-ossp Postgres extension is required to generate new UUIDs
        # for classifications as those are created from scratch
        # and Django handles defaults only in Python, not in the database.

        migrations.RunSQL(
            """
            CREATE EXTENSION IF NOT EXISTS "uuid-ossp";

            INSERT INTO documents_elementtype (id, corpus_id, slug, display_name, folder, default_view)
            SELECT id, corpus_id, 'region_' || name, name || ' region', FALSE, FALSE
            FROM documents_mlclass
            WHERE id in (SELECT DISTINCT ml_class_id FROM documents_region);

            INSERT INTO documents_element (id, corpus_id, type_id, name, zone_id, source_id, created, updated)
            SELECT
                region.id,
                ml_class.corpus_id,
                region.ml_class_id,
                ml_class.name || ' ' || ROW_NUMBER() OVER (
                    PARTITION BY region.element_id, region.ml_class_id ORDER BY region.id
                ),
                region.zone_id,
                region.source_id,
                CURRENT_TIMESTAMP,
                CURRENT_TIMESTAMP
            FROM documents_region region
            LEFT JOIN documents_mlclass ml_class
            ON (region.ml_class_id = ml_class.id);

            INSERT INTO documents_classification
                (id, element_id, source_id, ml_class_id, confidence, high_confidence, state)
            SELECT uuid_generate_v4(), id, source_id, ml_class_id, confidence, TRUE, 'pending'
            FROM documents_region;

            INSERT INTO documents_elementpath (id, element_id, path, ordering)
            SELECT
                uuid_generate_v4(),
                id,
                ARRAY[element_id],
                ROW_NUMBER() OVER (
                    PARTITION BY element_id, ml_class_id ORDER BY id
                ) - 1
            FROM documents_region
            WHERE element_id NOT IN (SELECT element_id FROM documents_elementpath);

            INSERT INTO documents_elementpath (id, element_id, path, ordering)
            SELECT
                uuid_generate_v4(),
                region.id,
                ep.path || element_id,
                ROW_NUMBER() OVER (
                    PARTITION BY element_id, region.ml_class_id, ep.path[array_length(ep.path, 1)] ORDER BY region.id
                ) - 1
            FROM documents_elementpath ep
            FULL JOIN documents_region region USING (element_id)
            WHERE region.id IS NOT NULL;
            """,
            reverse_sql=migrations.RunSQL.noop,
            elidable=True,
        ),
    ]
